#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""
R3 Process Single Manifest helper.
"""

import logging
from typing import List

from libs.constants import FIRST_STORED_RECORD_INDEX
from libs.context import Context
from libs.manifest_analyzer import ManifestAnalyzer
from libs.refresh_token import TokenRefresher
from libs.process_manifest_r3 import ManifestProcessor
from libs.traverse_manifest import ManifestTraversal
from libs.validate_schema import SchemaValidator
from libs.validation.validate_data_integrity import DataIntegrityValidator

logger = logging.getLogger()


class SingleManifestProcessor(object):

    def __init__(
        self,
        storage_url: str,
        payload_context: Context,
        data_integrity_validator: DataIntegrityValidator,
        manifest_processor: ManifestProcessor,
        schema_validator: SchemaValidator,
        token_refresher: TokenRefresher,
    ):
        """Init SingleManifestProcessor."""
        self.storage_url = storage_url
        self.payload_context = payload_context
        self.data_integrity_validator = data_integrity_validator
        self.manifest_processor = manifest_processor
        self.schema_validator = schema_validator
        self.token_refresher = token_refresher

    def _process_records(self, manifest_analyzer: ManifestAnalyzer) -> str:
        """
        Process each entity from entity queue created according to child-parent relationships
        between entities.
        Replace surrogate-keys of parents inside child entities with system-generated keys.

        :param manifest_analyzer: Object with proper queue of entities
        :yield: System-generated id of entity
        """
        for entity in manifest_analyzer.entity_queue():
            try:
                logger.debug(f"Process entity {entity}")
                entity.replace_parents_surrogate_srns()
                record_id = self.manifest_processor.process_manifest_records(
                    [entity.entity_info]
                )[FIRST_STORED_RECORD_INDEX]
                entity.system_srn = record_id
                yield record_id
            except Exception as e:
                logger.warning(f"Can't process entity {entity}")
                logger.error(e)
                manifest_analyzer.add_unprocessed_entity(entity)

    def process_manifest(self, manifest: dict) -> List[str]:
        """Execute manifest validation then process it.

        Execution steps:
        1) validata data integrity and prune to maintain dependency consistency
        2) initialize schema validator
        3) validate manifest file against common schema
        4) traverse manifest file and extract manifest entities
        5) validate extracted manifest entities
        6) create an ingestion queue according to child-parent relationships between entities
        7) process valid manifest entities one-by-one

        :param context: Airflow context
        :type context: dict
        :return: List of record ids.
        """
        record_ids = []
        self.data_integrity_validator.validate_manifest_data_integrity(manifest)
        manifest_schema = self.schema_validator.validate_common_schema(manifest)
        traversal = ManifestTraversal(manifest, manifest_schema)
        manifest_entities = traversal.traverse_manifest()
        logger.debug(f"entities count: {len(manifest_entities)}")
        valid_manifest_entities = self.schema_validator.validate_manifest(manifest_entities)
        logger.debug(f"valid entities count: {len(valid_manifest_entities)}")

        manifest_analyzer = ManifestAnalyzer(
            valid_manifest_entities,
            self.storage_url,
            self.token_refresher,
            self.payload_context
        )
        for record_id in self._process_records(manifest_analyzer):
            record_ids.append(record_id)
        logger.info(f"Processed ids {record_ids}")
        return record_ids
