#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""
R3 Process Manifest operator.
"""

import logging
from math import ceil
from typing import List, Tuple, TypeVar

from airflow.utils import apply_defaults
from airflow.models import BaseOperator, Variable
from jsonschema import SchemaError
from libs.context import Context
from libs.exceptions import EmptyManifestError, NotOSDUSchemaFormatError, \
    UploadFileError, GetSchemaError
from libs.source_file_check import SourceFileChecker
from libs.handle_file import FileHandler
from libs.refresh_token import AirflowTokenRefresher
from libs.process_manifest_r3 import ManifestProcessor
from libs.validate_schema import SchemaValidator
from libs.processors.single_manifest_processor import SingleManifestProcessor
from libs.search_client import SearchClient
from libs.validation.validate_data_integrity import DataIntegrityValidator
from requests import HTTPError

logger = logging.getLogger()

ManifestType = TypeVar("ManifestType", dict, List[dict])


class ProcessManifestOperatorR3(BaseOperator):
    """Operator to process manifest R3."""

    ui_color = '#dad5ff'
    ui_fgcolor = '#000000'

    @apply_defaults
    def __init__(self, previous_task_id: str = None, batch_number=3, *args, **kwargs):
        """Init base operator and obtain base urls from Airflow Variables."""
        super().__init__(*args, **kwargs)
        self.previous_task_id = previous_task_id
        self.batch_number = batch_number
        self.schema_service_url = Variable.get('core.service.schema.url')
        self.search_service_url = Variable.get('core.service.search.url')
        self.storage_url = Variable.get('core.service.storage.url')
        self.file_service_url = Variable.get('core.service.file.host')
        self.batch_count = int(Variable.get("core.batch_count", "3"))

    def _get_manifest_files_range(self, manifests: List[dict]) -> Tuple[int, int]:
        """
        Get start and end indexes of a manifest files slice to be processed within this task.
        :param manifests: A list of manifests
        :return: start index, end index
        """
        split_size = ceil(len(manifests) / self.batch_count)
        slice_start_index = (self.batch_number - 1) * split_size
        slice_end_index = self.batch_number * split_size
        return slice_start_index, slice_end_index

    def _process_manifest(
        self,
        single_manifest_processor: SingleManifestProcessor,
        manifest: ManifestType
    ) -> List[str]:
        """

        :param single_manifest_processor: Object to process a single manifest file.
        Processing includes validation against schemas, storing records enc.
        :param manifest: A single manifest file or a list of them.
        :return:
        """
        if isinstance(manifest, dict):
            record_ids = single_manifest_processor.process_manifest(manifest)

        elif isinstance(manifest, list):
            record_ids = []
            slice_start_index, slice_end_index = self._get_manifest_files_range(manifest)
            logger.debug(f"Start and indexes {slice_start_index}:{slice_end_index}")
            for single_manifest in manifest[slice_start_index:slice_end_index]:
                logger.debug(f"processing {single_manifest}")
                try:
                    record_ids.extend(single_manifest_processor.process_manifest(single_manifest))
                except (UploadFileError, HTTPError, GetSchemaError, SchemaError) as e:
                    logger.error(f"Can't process {single_manifest}")
                    logger.error(e)
                    continue
        else:
            raise NotOSDUSchemaFormatError(
                f"Manifest {manifest} must be either not empty 'list' or 'dict'")

        return record_ids

    def _get_manifest_data(self, context: dict, execution_context: dict) -> ManifestType:
        if self.previous_task_id:
            previous_task_value = context["ti"].xcom_pull(task_ids=self.previous_task_id,
                                                          key="return_value")
            if previous_task_value:
                manifest_data = previous_task_value["manifest"]
            else:
                manifest_data = execution_context["manifest"]
        else:
            manifest_data = execution_context["manifest"]
        return manifest_data

    def execute(self, context: dict):
        """Execute manifest validation then process it.

        Get a single manifest file or a list of them.
        If it is a list, calculate which range (slice) of manifest files must be processed and then
        process this range one by one.

        :param context: Airflow context
        :type context: dict
        """
        execution_context = context["dag_run"].conf["execution_context"]
        payload_context = Context.populate(execution_context)
        token_refresher = AirflowTokenRefresher()
        file_handler = FileHandler(self.file_service_url, token_refresher, payload_context)
        search_client = SearchClient(self.search_service_url, token_refresher, payload_context)
        source_file_checker = SourceFileChecker()

        data_integrity_validator = DataIntegrityValidator(search_client=search_client)

        manifest_processor = ManifestProcessor(
            storage_url=self.storage_url,
            file_handler=file_handler,
            token_refresher=token_refresher,
            context=payload_context,
            source_file_checker=source_file_checker,
        )
        validator = SchemaValidator(
            self.schema_service_url,
            token_refresher,
            payload_context
        )
        single_manifest_processor = SingleManifestProcessor(
            storage_url=self.storage_url,
            payload_context=payload_context,
            data_integrity_validator=data_integrity_validator,
            manifest_processor=manifest_processor,
            schema_validator=validator,
            token_refresher=token_refresher,
        )

        manifest_data = self._get_manifest_data(context, execution_context)
        logger.debug(f"Manifest data: {manifest_data}")

        if not manifest_data:
            raise EmptyManifestError(
                f"Data {context['dag_run'].conf} doesn't contain 'manifest field'")
        record_ids = self._process_manifest(single_manifest_processor, manifest_data)
        logger.info(f"Processed ids {record_ids}")
        context["ti"].xcom_push(key="record_ids", value=record_ids)
